@@ -19,6 +19,7 @@
 #include "uni.h"
 #include "../fab.h"
 #include "../dispatch.h"
+#include "tcpswc.h"
 
 
 #define COMPONENT "FAB(tcp)"
@@ -35,7 +36,7 @@
 typedef struct _table {
     struct _link *out;	/* output port */
     uint16_t in_vpi;	/* input VPI */
-    int	in_vci;		/* input VCI */
+    int	in_vci;		/* input VCI (may be ATM_VCI_ANY) */
     uint16_t out_vpi;	/* output VPI */
     uint16_t out_vci;	/* output VCI */
     struct _table *next;
@@ -60,6 +61,7 @@ typedef struct _fab {
 static CALL *calls = NULL;
 static LINK *links = NULL;
 static void (*notify)(int number,int up);
+static int s_control = -1;
 
 
 static LINK *find_link(int id)
@@ -314,6 +316,74 @@ static int check_ci(struct sockaddr_atmpvc *pvc)
 }
 
 
+static int msg_handler(void *buf,int len,void *user)
+{
+    TCPSWC_MSG *msg = buf;
+    LINK *lnk;
+    TABLE *entry;
+    int i;
+
+    if (len != sizeof(TCPSWC_MSG))
+	diag(COMPONENT,DIAG_FATAL,"msg_handler: bad length (%d != %d)",len,
+	  sizeof(TCPSWC_MSG));
+    switch (msg->type) {
+	case tmt_get:
+	    i = msg->n;
+	    for (lnk = links; lnk; lnk = lnk->next) {
+		for (entry = lnk->table; i && entry; entry = entry->next) i--;
+		if (!i && entry) break;
+	    }
+	    if (!lnk || !entry) {
+		msg->n = -ENOENT;
+		break;
+	    }
+	    memset(&msg->in,0,sizeof(msg->in));
+	    msg->in.sap_family = AF_ATMPVC;
+	    msg->in.sap_addr.itf = lnk->id;
+	    msg->in.sap_addr.vpi = entry->in_vpi;
+	    msg->in.sap_addr.vci = entry->in_vci;
+	    memset(&msg->out,0,sizeof(msg->out));
+	    msg->out.sap_family = AF_ATMPVC;
+	    msg->out.sap_addr.itf = entry->out->id;
+	    msg->out.sap_addr.vpi = entry->out_vpi;
+	    msg->out.sap_addr.vci = entry->out_vci;
+	    memset(&msg->qos,0,sizeof(msg->qos));
+	    break;
+	case tmt_add:
+	    msg->n = add_entry(&msg->in,&msg->out);
+	    break;
+	case tmt_del:
+	    msg->n = del_entry(&msg->in,&msg->out);
+	    break;
+	default:
+	    diag(COMPONENT,DIAG_FATAL,"msg_handler: unknown message type %d",
+	      msg->type);
+    }
+    return sizeof(TCPSWC_MSG);
+}
+
+
+static void control_msg(int sock,void *dummy)
+{
+    TCPSWC_MSG msg;
+
+    if (un_reply(sock,&msg,sizeof(msg),&msg_handler,NULL) < 0)
+	diag(COMPONENT,DIAG_ERROR,"un_reply: %s",strerror(errno));
+}
+
+
+void fab_option(const char *name,const char *value)
+{
+    if (strcmp(name,"control"))
+	diag(COMPONENT,DIAG_FATAL,"unrecognized fabric option \"%s\"",name);
+    if (s_control != -1)
+	diag(COMPONENT,DIAG_FATAL,"control channel is already set");
+    s_control = un_create(value,0600);
+    if (s_control < 0)
+	diag(COMPONENT,DIAG_FATAL,"un_create: %s",strerror(errno));
+}
+
+
 void fab_start(void (*port_notify)(int number,int up))
 {
     struct sockaddr_in addr;
@@ -330,6 +400,7 @@ void fab_start(void (*port_notify)(int number,int up))
     if (listen(s_listen,5) < 0)
 	diag(COMPONENT,DIAG_FATAL,"listen: %s",strerror(errno));
     dsp_fd_add(s_listen,new_link,NULL);
+    if (s_control >= 0) dsp_fd_add(s_control,control_msg,NULL);
 }
 
 
